
(require "zpackage" "zpackage.lisp")
(in-package "Z")

(require "utilities")
(provide 'operators)

(defun domain (a) (mapcar #'car a))

(defun range  (a) (mapcar #'cdr a))

(defun size (x) (length x))

(defun inv (rel) (pairlis (range rel) (domain rel)))

(defun in (a b) (member a b :test #'equal))

(defun notin (a b) (not (in a b)))

(defun subeq (a b) (subsetp a b :test #'equal))

(defun seteq (a b) (and (subeq a b) (subeq b a)))

(defun equal* (a b)
  (or (equal a b)
      (and (consp a) (consp b) (seteq a b))))

(defun image (rel set) 
  (cond ((not (consp rel)) nil)
	((member (caar rel) set) (cons (cdar rel) (image (cdr rel) set)))
	(t (image (cdr rel) set))))

(defun dres (domain rel)
  (cond ((not (consp rel)) rel)
	((member (caar rel) domain) (cons (car rel) (dres domain (cdr rel))))
	(t (dres domain (cdr rel)))))

(defun rres (rel range)
  (cond ((not (consp rel)) rel)
	((member (cdar rel) range) (cons (car rel) (rres (cdr rel) range)))
	(t (rres (cdr rel) range))))

(defun ndres (domain rel)
  (cond ((not (consp rel)) rel)
	((member (caar rel) domain) (ndres domain (cdr rel)))
	(t (cons (car rel) (ndres domain (cdr rel))))))

(defun nrres (rel range)
  (cond ((not (consp rel)) rel)
	((member (cdar rel) range) (nrres (cdr rel) range))
	(t (cons (car rel) (nrres (cdr rel) range)))))

(defun oplus (a b)
  "A and B are relations.  The resulting relation relates everything
in B as before and deletes pairs whose first el is in the domain of
B from A."
  (union (ndres (domain b) a) b))

(defun mapsto (a b) (cons a b))

(defun parens (a) a)

(defmacro set* (&rest l) (cons 'list l))

(defmacro set-comp (sch &optional expr)
  (let (var values typ pred)
    (cond ((and expr (match sch (schematext ((basic-decl nil . x)))
			    (format t "{ ~s @ ~s} not handled" (car x) expr)))
	   ;; `(basic-set ',x nil ',expr)))
	   nil)
	  ((match sch (schematext ((basic-decl nil . x)) preds)
		  (format t "{ ~s | ~s @ ~s} not handled" (car x) preds expr))
	   ;; `(basic-set ',x ',preds ',expr)
	   nil)
	  ((match sch (schematext ((basic-decl nil . x)))
		  (setq values x))
	   `(quote ,values))
	  ((match sch (schematext ((basic-decl type x)) preds)
		  (progn (setq var x)
			 (setq typ type)
			 (setq pred preds)))
	   `(create-set ',var ',typ ',pred ',expr))
	  ((match sch (schematext ((basic-decl type x)))
		  (progn (setq var x)
			 (setq typ type)))
	   `(create-set ',var ',typ nil ',expr))
	  (t (cons 'list sch)))))

;; (defun basic-set (cands preds expr)
;;   (sloop for x in cands
;; 	 when (or (null preds) (schema-eval `(let ((,x ',x)) ,preds)))
;; 	 collect (if expr
;; 		     (schema-eval `(let ((,x ',x)) ,expr))
;; 		     x))))

(defvar null nil)

(defun set-comprehension (st &optional expr)
  ;; st is schema-text
  (cond ((match st (schematext ((basic-decl type x)) preds)
		(create-set x type preds expr)))
	((match st (schematext ((basic-decl type x)))
		(create-set x type nil expr)))
	(t (format t "~%At the moment we only handle trivial set comprehensions, e.g.,
 { x : Type | f(x) }")
	   nil)))

(defun create-set (var type preds expr)
  (let ((cand (many-values-of type)))
    (sloop for x in cand 
	   when (or (null preds) (schema-eval `(let ((,var ',x)) ,preds)))
	   collect (if expr
		       (schema-eval `(let ((,var ',x)) ,expr))
		       x))))

(defvar *integer-default-set-comprehension* (list -5 -4 -3 -2 -1 0 1 2 3 4 5))

(defvar *types* nil "Global type alist")

(defun many-values-of (type)
  ;; If type is simple and finite, return complete set
  ;; If type is N, return -5 ... 5
  ;; If type is complex, build recursively.
  (cond ((symbolp type)
	 (cond ((member type '(z integer integerp))
		*integer-default-set-comprehension*)
	       ((assoc type *types*)
		(let ((struct (cdr (assoc type *types*))))
		      (let* ((basetype (ztype-basetype struct))
			     (constants (ztype-constants struct)))
			(cond (constants constants)
			      ((eq basetype type) nil)
			      (basetype (many-values-of basetype))
			      (t (error (format nil "Ill-formed typedef for ~s" type)))))))
	       ((getv type))
	       (t (error (format nil "Reference to undefined type or function, ~s" type)))))
	((consp type)
	 (cond ((match type (pfun x y) 
		       (pairup (many-values-of x) (many-values-of y))))
	       ((match type (tfun x y)
		       (pairup-total (many-values-of x) (many-values-of y) (many-values-of y))))
	       ((match type (cross . l)
		       (multiply* (sloop for x in l
					 collect (many-values-of x)))))
	       ((member (car type) '(powerset p1 f f1 seq))
		(powerset (cadr type)))
	       (t (error "Bad type specification ~a" type))))
	(t (error "Bad type specification ~a" type))))

(defun pairup (a b)
  (cond ((or (atom a) (atom b)) nil)
	(t (cons (cons (car a) (car b)) (pairup (cdr a) (cdr b))))))

(defun pairup-total (a b c)
  (cond ((atom a) nil)
	((atom b)
	 (if (atom c)
	     (warning "Bad pairup, 2nd (~a) and 3rd (~a) args not consp:" b c)
	     (pairup-total a c c)))
	(t (cons (cons (car a) (car b)) (pairup-total (cdr a) (cdr b) c)))))

(defun multiply (a b) (multiply* (list a b)))

(defun multiply* (l)
  (cond ((null l) nil)
	((null (cdr l)) (mapcar #'list (car l)))
	(t (let ((rest (multiply* (cdr l))))
	     (sloop for x in (car l)
		    append (sloop for y in rest
				   collect (cons x y)))))))

(defun remove-vars (l vars)
  ;; L is a list of (name . type)
  (cond ((symbolp l) l)
	((member (car l) vars) (remove-vars (cdr l) vars))
	(t (cons (car l) (remove-vars (cdr l) vars)))))

(defun hide (s vars)
  (cond ((or (schemap s) (schema-p s))
	 (make-schema :name (schema-name s) 
		      :kind (schema-kind s)
		      :decls (schema-decls s)
		      :locals (remove-vars vars (schema-locals s))
		      :vars (remove-vars vars (schema-vars s))
		      :primes (remove-vars vars (schema-primes s))
		      :schemas (schema-schemas s)
		      :signature (remove-vars vars (schema-signature s))
		      :preds (schema-preds s)
		      :functions (schema-functions s)
		      :input (remove-vars vars (schema-input s))
		      :output (remove-vars vars (schema-output s))
		      ;; ???
		      :exists (schema-exists s)
		      ))
	(t (set-difference s vars :test #'equal))))
