
(require "zpackage" "zpackage.lisp")
(in-package "Z")

(provide "structures")

;; ------------------------------------------------------------------------
;; Structures

(defstruct (schema (:predicate schemap))
   name      ; NIL || symbol || (DELTA symbol) || (XI symbol) || (symbol . vars)
   gen-formals ; (symbol)*
   kind      ; a schema type: SCHEMA | GENERIC-SCHEMA | GENERIC-CONSTANT
   next      ; forall undecorated x with a primed variant x', next = x'
   prev      ; forall primed x' with an undecorated parent x, prev = x
   decls     ; expr
   preds     ; expr
   type      ; cross product of types of vars.
   schemas   ; (schema-name*)
   locals    ; ((symbol . type)*) - vars from decls
   bindings  ; ((localsymbol . value)*)
   vars      ; ((symbol . type)*) - union of local and vars of schemas
   functions ; ((symbol . type)*)
   defns     ; ((symbol . defn)*)
   ;types    ; a list of types declared in schema, sorted by name.
   signature ; a list of (var.basetype) sorted by var name.
             ; Thus signature equality just uses equal.
   input   ; ((symbol . type)*)  - (subset input vars)
   output  ; ((symbol . type)*)  - (subset output vars)
   primes  ; ((symbol . type)*)  - Primed vars, (subset primes vars)
   exists  ; ((symbol . type)*)  - (empty (intersection exists vars))  ???
   exec    ; The function to call, with arg types. (fn t1 t2 .. tn).
           ; Roughly
           ;  (let ((!1 e1) ... (!n en)) (mv !1 ... !n))
           ;   ST   '(equal !i ei) in preds
           ;     or '(REL !i f) in preds
           ;        and ei = (choose !i !i.TYPE (REL !i f))
   updated ;; Indicates that the x' vars have been updated and not
           ;; yet copied into the x vars. 
)

;; Supported types are (pfun type type), (tfun type type),
;; (powerset type), (seq type), (cross . type*),
;  (function id) - where (fboundp id) and (match (signature id) '((x) BOOLEAN))
;; id - where (assoc id *types*).

(defstruct (ztype (:predicate ztypep))
                ; z             base           free
  name          ; z             symbol         symbol
  kind          ; integer       BASIC          FREE
  basetype      ; Z
  fun           ; integerp
  firstchoice   ; 0             0              a
  constants     ; NIL           NIL            a, b, c
  constructors  ; NIL           NIL            ((cons n n))
  )

(defstruct (zglobal (:predicate zglobalp))
                ; N             base           free
  name          ; N             symbol         symbol
  type          ; 
  value         ;
  pred          ; any restriction on value
  )

(defstruct (zconstructor (:predicate zconstructorp))
  name
  args
  typename)

(defstruct (zfunction (:predicate zfunctionp))
  name
  args
  type
  basetype
  preds)
