
(require "zpackage" "zpackage.lisp")
(in-package "Z")

;; (require "sloop")
(use-package "SLOOP")

(provide "utilities")

(eval-when (load eval compile)

(defmacro f-l (vars body) `(function (lambda ,vars ,body)))
(defmacro ffn-args (x) `(cdr ,x))
(defmacro ffn-symb (x) `(car ,x))
(defmacro fquotep (x) `(eq (car ,x) (quote quote)))
(defmacro cons-term (x y) `(cons ,x ,y))

)

(defmacro pack (&rest l)
  `(intern (concatenate 'string
			,@(mapcar
			   (function (lambda (x) (list 'symbol-name x)))
			   l))))

(defun pack2 (l)
  (if (characterp l)
      (string l)
      (if (symbolp l)
	  ""
	  (concatenate 'string (string (car l)) (pack2 (cdr l))))))

(defun pack* (chars)
  ;; This hack is about 6 times faster (uncompiled) than calling pack2.
  (intern (format nil "~{~c~}" chars)))

(defun check-matches (form pattern)
  ;; pattern may be (a . b), or as a result, b.
  (cond ((symbolp form) (eq form pattern))
	;; Matches the list, form.
	((symbolp pattern) t)
	((check-match (car form) (car pattern))
	 (check-matches (cdr form) (cdr pattern)))
	(t nil)))

(defun check-match (form pattern)
  ;; a pattern is a constant (number, char, T, NIL), a variable (symbolp)
  ;; or a consp whose car is treated as a constant and whose cdr is a 
  ;; list of patterns.
  (cond ((equal form pattern) t)
	((equal pattern '&) t)
	((or (numberp pattern) 
	     (eq pattern t)
	     (eq pattern nil)
	     (characterp pattern)) nil)
	((symbolp pattern) t)
	((not (consp form)) nil)
	((and (consp (car pattern))
	      (consp (car form)))
	 (and (check-match (car form) (car pattern))
	      (check-matches (cdr form) (cdr pattern))))
	((equal (car form) (car pattern))
	 (check-matches (cdr form) (cdr pattern)))
	(t nil)))

(defun bind-match (root pattern &optional listp)
  ;; a pattern is a constant (number, char, T, NIL), a variable (symbolp)
  ;; or a consp whose car is treated as a constant and whose cdr is a 
  ;; list of patterns.
  (cond (listp
	 (cond ((consp pattern)
		(append (bind-match (list 'car root) (car pattern))
			(bind-match (list 'cdr root) (cdr pattern) t)))
	       ((null pattern) nil)
	       (t (bind-match root pattern))))
	((or (equal pattern '&)
	     (numberp pattern) 
	     (eq pattern t)
	     (eq pattern nil)
	     (characterp pattern)) nil)
	((symbolp pattern) (list (list pattern root)))
	((not (consp pattern)) nil)
	((consp (car pattern))
	 (append (bind-match (list 'car root) (car pattern) nil)
		 (bind-match (list 'cdr root) (cdr pattern) t)))
	((consp pattern)
	 (bind-match (list 'cdr root) (cdr pattern) t))))

(defmacro match (form pattern &optional (value nil valuep))
  (if (not valuep)
      `(check-match ,form ',pattern)
      `(if (check-match ,form ',pattern)
	   (let ,(bind-match form pattern)
	     ,value))))

(defmacro bind (form pattern value)
  ;; Presumes match has been done.
  `(let ,(bind-match form pattern) ,value))

(defun addbinding (l field value)
  (let ((pair (assoc field l)))
    (if pair
	(if (equal (cdr pair) value)
	    l
	    (cons (cons field value) (remove pair l :test #'equal)))
	(cons (cons field value) l))))

(defmacro setassoc (l field value &optional warn)
  (if warn
      `(let ((new ,value))
	 (if (assoc ,field ,l)
	     (progn (if (equal (cdr (assoc ,field ,l)) new)
			nil
			(progn (format t ,warn ,field)
			       (setf (cdr (assoc ,field ,l)) new))))
	     (setq ,l (cons (cons ,field new) ,l))))  
      `(let ((new ,value))
	 (if (assoc ,field ,l)
	     (setf (cdr (assoc ,field ,l)) new)
	     (setq ,l (cons (cons ,field new) ,l))))))

(defun in-p (var p)
  (cond ((symbolp p) (eq var p))
	((in-p var (car p)) t)
	(t (in-p var (cdr p)))))

(defmacro warning (&rest l) `(progn (format t ,@l) nil))

(defmacro retain (x pred &rest l) `(remove-if-not ,pred ,x ,@l))


