(require "zpackage" "zpackage.lisp")
(in-package "Z")

(require "utilities")
(require "structures")

(provide 'zlisp-events)

;; Events from types

(defun def-type-events (types) (mapc (function ev-type) types))

; ------------------------------------------------------------
; EV-TYPE - assert a base type.
;  There is nothing more to say.  We can't even say for sure 
;  that they are disjoint from other types.
; ------------------------------------------------------------

(defun ev-type (type)
  (if (consp type) (setq type (cdr type)))
  (let ((name (ztype-name type)))
  `(encapsulate ((,name (x) t))
		(local (defun ,name (x) (rationalp x)))
		(defthm ,(pack name '-type) 
		  (or (not (,name x)) (,name x))))))


;; (defun def-schema-box-events (schema)
;;   (let ((name (schema-name schema))
;; 	vars (schema-box-vars decl-part))
;; 	(axioms (andlist (append (decl-axioms decl-part) where))))
;;     `(,@(decl-type-generators (no-duplicates (new-schema-types decl-part) :test #'equal))
;;       (defun ,(ztype-fn schema-name) ,vars
;; 	,(andlist (decl-types decl-part)))
;;       (defun ,(axiom-fn schema-name) ,vars ,axioms)
;;       (setq *schemas* (cons (cons schema-name
;; 				  (make-schema :name ',schema-name
;; 					       :vars ',vars
;; 					       ;; :exists ',existentials
;; 					       :preds '(,(axiom-fn schema-name) ,vars)))
;; 			    *schemas*))
;;       (defun ,schema-name (,*schema-parameter*)
;; 	(and (all-boundp (quote ,vars) ,*schema-parameter*)
;; 	     (,(ztype-fn schema-name)
;; 	      ,@(sloop for var in vars
;; 		       collect `(binding (quote ,var) ,*schema-parameter*)))
;; 	     (,(axiom-fn schema-name)
;; 	      ,@(sloop for var in vars
;; 		       collect `(binding (quote ,var) ,*schema-parameter*)))))))

(defvar *schema-parameter* 's)

(defun defun-schema-events (s)
  "Meant to be used to create ACL2 defuns and theorems."
  (let* ((name (schema-name s))
	 (decl-part (schema-decls s))
	 (vars (schema-vars s))
	 (varnames (mapcar (function car) vars))
	 (axioms (schema-preds s)))
    `(,@(decl-type-generators (schema-types s))
	(defun ,(ztype-fn name)  ,varnames ,(andlist (decl-types decl-part)))
	(defun ,(axiom-fn name) ,varnames ,axioms)
	(defun ,name (,*schema-parameter*)
	  (and (all-boundp (quote ,varnames) ,*schema-parameter*)
	       (,(ztype-fn name)
		,@(sloop for var in varnames
			 collect `(binding (quote ,var) ,*schema-parameter*)))
	       (,(axiom-fn name)
		,@(sloop for var in varnames
			 collect `(binding (quote ,var) ,*schema-parameter*))))))))

(defun decl-types (decl-part)
  (sloop for decl in decl-part
	   append (cond ((match decl (basic-decl nil . schemas)) nil)
			((match decl (basic-decl type . vars)
				(sloop for var in vars
				       collect `(,(ztype-function type) ,var))))
			((match decl (schema x)
				(schexp-types x)))
			((match decl (DELTA x)) nil)
			((match decl (XI x)) nil)
			(t (warning "Bad decl ~a:" decl)))))

(defmacro defun-axiom-events (a)
  ;; name types term hints
  (let* ((s (eval a))
	 (name (car s))
	 ;(types  (schema-types s))
	 (vars   (schema-vars s))
	 (preds  (schema-preds s)))
  `(defthm ,(theorem-fn name)
     (implies ,(conjunct (make-global-assertions vars))
	      ,preds))))

; ------------------------------------------------------------
; DEF-LEMMA
; ------------------------------------------------------------

; Example
; 
; (def-lemma add-birthday-known-fact (rewrite)
;   (implies (schema add-birthday)
; 	   (equal known (append known (list name?)))))

; REMARK There should probably be some consolidation of the
;        schema expressions and the above notation. Does
;        a schema expression allow schema references to be
;        intermixed with arbitrary terms?

(defmacro def-lemma (name types term &optional hints)
  `(progn
    (quote (,@(def-lemma-events name types term hints)))))

(defun def-lemma-events (name types term hints)
  `((prove-lemma ,name ,types ,(interpret-zterm term) ,hints)))

;; base-type fn

(defvar *types* nil)

(defun base-type-fn (type term)
  ;; Type is an expanded Basetype.
  (cond ((equal type 'integer) `(integerp ,term))
	((equal type 'natural) `(naturalp ,term))
	((symbolp type) (if (assoc type *types*)
			    (base-type-fn (ztype-basetype (cdr (assoc type *types*))) term)
			    (warning "~%~S is not a type" type)))
	((consp type)
	 (cond ((member (car type) '(cross rel pfun fun pinj inj psurj surj bij ffun finj))
		(cons 'and 
		      (let ((n -1))
			(mapcar (f-l (el)
				     (base-type-fn el `(nth ,(incf n) ,term)))
				(cdr type)))))
	       ((member (car type) '(powerset p1 f f1))
		`(every z ,term ,(base-type-fn (cadr type) 'z)))
	       ((member (car type) '(set))
		;; (set var type suchthat pred)
		`(every z ,term ,(base-type-fn (nth 4 type) 'z)))
	       (t (error "Bad schema expression ~a" type))))
	(t (warning "~%~S is not a type" type))))


