
 /* ZTC Parser for Z ascii input.  Produces a file os sexprs. */

%{

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h> 
#include "scansubs.h"
#include "semsubs.h"
#include "types.h"
/* #define  YYDEBUG 1 */
/* #define YYERROR_VERBOSE 1 */
#define  YYDEBUG 1
#define YYERROR_VERBOSE 1

%}

%union{
 char  *charptr;
 CELL  *treeptr;
 SYMBOLS *symptr;
 }

%token <symptr>  TOK_IDENTIFIER

%token <charptr> TOK_STRING_LITERAL TOK_INTEGER_LITERAL 

%token <charptr> TOK_VBAR TOK_SEP TOK_SIZE
%token <charptr> TOK_IF TOK_THEN TOK_ELSE
%token <charptr> TOK_AND TOK_AXIOM TOK_BAG  
%token <charptr> TOK_BEGIN TOK_DISJOINT TOK_DIV TOK_END 
%token <charptr> TOK_EXISTS1 TOK_FALSE TOK_FORALL
%token <charptr> TOK_GENERIC TOK_GLOBAL TOK_INBAG TOK_ID
%token <charptr> TOK_IS TOK_ISEQ TOK_LAMBDA
%token <charptr> TOK_MOD TOK_NOT TOK_NOTIN TOK_PARTITION 
%token <charptr> TOK_PRE TOK_OR TOK_SCHEMA TOK_RAN TOK_DOM
%token <charptr> TOK_SEQ TOK_SEQ1 TOK_SPEC TOK_INIT TOK_DATA TOK_SUB 
%token <charptr> TOK_SUBEQ TOK_THETA TOK_TRUE TOK_UNIQUE TOK_WHERE 
%token <charptr> TOK_CHANGED TOK_DELTA TOK_F TOK_FALSE 
%token <charptr> TOK_F1 TOK_INTERSECTION TOK_N TOK_N1 TOK_P TOK_P1 TOK_TRUE
%token <charptr> TOK_UNCHANGED TOK_UNION TOK_XI TOK_Z TOK_LET
%token <charptr> TOK_REL TOK_PFUN TOK_TFUN TOK_PINJ TOK_ING TOK_PSURJ
%token <charptr> TOK_SURJ TOK_BIJ TOK_FFUN TOK_FINJ TOK_LANGLE TOK_RANGLE
%token <charptr> TOK_EXISTS TOK_EQUAL TOK_IN TOK_NEQ TOK_LTE TOK_GTE TOK_CROSS
%token <charptr> TOK_INV TOK_PLUS TOK_STAR TOK_LIMG TOK_RIMG TOK_LBAG TOK_RBAG
%token <charptr> TOK_DOT TOK_INJ TOK_EQUIV TOK_DCAT TOK_EMPTY TOK_INREL
%token <charptr> TOK_FREE_TYPE

/*  Precedence - This only applies when these literals are the first non-term 
    in a rule.  (Or if you use %prec.)  */

%left  TOK_DRES TOK_RRES TOK_NDRES TOK_NRRES
%left  TOK_OPLUS
%left  '*' TOK_DIV TOK_CAP TOK_MOD TOK_COMP TOK_CIRC TOK_FILTER
%left  '+' '-' TOK_CUP '\\' TOK_CAT TOK_UPLUS
%left  TOK_UPTO
%left  TOK_MAPSTO

%left  TOK_AND
%left  TOK_OR
%left  TOK_NOT
%right TOK_IMPLIES
%left  TOK_IFF
%left  TOK_PROJECT
%left  TOK_HIDE
%left  TOK_SEMI

%type <charptr> specword globax in_gen pre_gen spec_label 
%type <treeptr> deco deco_list decorated_ident 
%type <treeptr> descriptions initial assign inits init_label init data_init label_end
%type <treeptr> specification paragraphs paragraph error 
%type <treeptr> ident idents decl_part axiom_part schema_header schema_header 
%type <treeptr> basic_decl def_lhs branch branches schema_exp schema_exp1 schema_text 
%type <treeptr> schema_or_var_ref decl_names let_def let_defs renamings renaming
%type <treeptr> predicate predicate1 expr expr0 expr1 expr2 expr3 expr_list
%type <treeptr> gen_actuals gen_formals set_expr integer_literal string_literal

%%

descriptions: specification initial { $$ = NULL; }
 | specification { $$ = NULL; }
;
initial: init_label inits TOK_SEP TOK_END TOK_INIT
  { $$ = NULL; printf("\n");
    if (fail_flag == 1) exit(1); }
 | init_label inits TOK_END TOK_INIT
  { $$ = NULL; printf("\n");
    if (fail_flag == 1) exit(1); }
 |error        { exit(3); }
;
inits: inits TOK_SEP init { $$ = NULL; }  
 | init { $$ = NULL; }  
;
init: data_init { $$ = NULL }
 | assign { $$ = NULL }
;
data_init: TOK_DATA ident
  { printf("\n(schema-data "); topwritetree($2); printf("\n"); } inits label_end TOK_DATA
  { $$ = NULL; printf(")\n"); }  
;
label_end: TOK_SEP TOK_END { $$ = NULL; }
 | TOK_END { $$ = NULL; }
;
assign: ident TOK_EQUAL expr 
{ topwritetree(cons_terminal("id-data",cons($1,list1($3)))); }
;
specification: spec_label paragraphs TOK_SEP TOK_END spec_label
  { $$ = NULL; printf("\n");
    if (fail_flag == 1) exit(1); }
 | spec_label paragraphs TOK_END spec_label
  { $$ = NULL; printf("\n");
    if (fail_flag == 1) exit(1); }
 |error        { exit(3); }
;
spec_label: TOK_SPEC TOK_SEP { $$ = NULL; }
 | TOK_SPEC { $$ = NULL; }
;
init_label: TOK_INIT TOK_SEP { $$ = NULL; }
 | TOK_INIT { $$ = NULL; }
;
paragraphs:  paragraph
  { topwritetree($1); $$ = NULL; printf("\n"); }  
 | paragraphs TOK_SEP paragraph { topwritetree($3); $$ = NULL; printf("\n"); }
 | error TOK_END specword TOK_SEP
  { topwritetree(cons_terminal("ERROR", NULL));
    fail_flag = 1;
    fprintf(stderr, "\n;Skipping to next form from line %d.",yylineno+1);}
    paragraphs 
 | error TOK_SEP
  { topwritetree(cons_terminal("ERROR", NULL));
    fail_flag = 1;
    fprintf(stderr, "\n;Skipping to next form from line %d.",yylineno+1);}
    paragraphs 
;
specword: TOK_AXIOM | TOK_GENERIC | TOK_GLOBAL | TOK_SCHEMA
;
paragraph:  '[' idents ']' { $$ = cons_terminal("base-type-decl",$2); } 
  | TOK_GLOBAL decl_part TOK_WHERE axiom_part TOK_END TOK_GLOBAL
   { $$ = cons_terminal("axiom",cons($2,list1($4))); }
  | TOK_GLOBAL decl_part TOK_AXIOM axiom_part TOK_END TOK_GLOBAL
   { $$ = cons_terminal("axiom",cons($2,list1($4))); }
  | TOK_AXIOM decl_part TOK_WHERE axiom_part TOK_END TOK_AXIOM
   { $$ = cons_terminal("axiom",cons($2,list1($4))); }
  | TOK_GLOBAL decl_part TOK_END TOK_GLOBAL
   { $$ = cons_terminal("axiom",cons($2,cons_terminal("nil",NULL))); }
  | TOK_AXIOM axiom_part TOK_END TOK_AXIOM
   { $$ = cons_terminal("axiom",cons_terminal("nil",cons($2,NULL))); }
  | globax decl_part TOK_END globax
   { $$ = cons_terminal("axiom",cons($2,cons_terminal("nil",NULL))); }
 | TOK_SCHEMA schema_header decl_part TOK_WHERE axiom_part  TOK_END TOK_SCHEMA
  { $$ = cons_terminal("schema",cons($2,cons($3, list1($5)))); }
 | TOK_SCHEMA schema_header decl_part TOK_END TOK_SCHEMA
  { $$ = cons_terminal("schema",cons($2,cons($3, NULL))); }
 | TOK_GENERIC gen_formals decl_part TOK_WHERE axiom_part  TOK_END TOK_GENERIC
  { $$ = cons_terminal("gendef",cons($2,cons($3, list1($5)))); }
 | TOK_GENERIC gen_formals decl_part TOK_END TOK_GENERIC
  { $$ = cons_terminal("gendef",cons($2,cons($3, NULL))); }
 | schema_header TOK_IS schema_exp 
  { $$ = cons_terminal("schemadef",cons($1,list1($3))); }
 | def_lhs TOK_EQUIV schema_exp 
  { $$ = cons_terminal("genequiv",cons($1,list1($3))); }
 | ident TOK_FREE_TYPE branches
  { $$ = cons_terminal("free-type",cons($1,list1($3))); } 
 | predicate { $$ = cons_terminal("axiom",cons_terminal("nil",cons(list1($1),NULL))); }
;
globax: TOK_GLOBAL { $$ = "global";} | TOK_AXIOM { $$ = "axiom";}
;
schema_header:
   ident gen_formals { $$ = cons($1,$2); }
 | ident { $$ = cons($1,NULL); }
;
 /* decl_name: ident { $$ = $1; };
 /* leaving out '(' op_name ')' for now */

decl_part:
   basic_decl { $$ = list1($1); }
 | decl_part TOK_SEP basic_decl { $$ = append($1,list1($3)); }
;
axiom_part:
   predicate { $$ = list1($1); }
 | axiom_part TOK_SEP predicate { $$ = append($1,list1($3)); }
;
def_lhs:
   ident gen_formals { $$ = cons_terminal("lhs",cons($1,$2)); }   
 | ident in_gen deco_list ident
  { $$ = cons_terminal("lhs",cons(cons_terminal($2,$3),cons($1,list1($4)))); }   
 | ident in_gen ident
  { $$ = cons_terminal("lhs",cons_terminal($2,cons($1,list1($3)))); }   
 | ident { $$ = cons_terminal("lhs",cons($1,NULL)); }   
 | pre_gen deco_list ident
  { $$ = cons_terminal("lhs",cons(cons_terminal($1,$2),list1($3))); }
 | pre_gen ident
  { $$ = cons_terminal("lhs",cons_terminal($1,list1($2))); }
;
deco_list:
   deco { $$ = cons($1,NULL); }
 | deco_list deco { $$ = append($1,cons($2,NULL)); }
;
deco:
   '\''  { $$ = cons_terminal("prime",NULL); }
 | '?'   { $$ = cons_terminal("input",NULL); }
 | '!'   { $$ = cons_terminal("output",NULL); }
 | '_' integer_literal { $$ = cons_terminal("subscript",list1($2)); }
;
in_gen:
      TOK_REL    { $$ = "rel";}
    | TOK_PFUN   { $$ = "pfun";}
    | TOK_TFUN    { $$ = "tfun";}
    | TOK_PINJ   { $$ = "pinj";}
    | TOK_INJ    { $$ = "inj";}
    | TOK_PSURJ  { $$ = "psurj";}
    | TOK_SURJ   { $$ = "surj";}
    | TOK_BIJ    { $$ = "bij";}
    | TOK_FFUN   { $$ = "ffun";}
    | TOK_FINJ   { $$ = "finj";}
;
pre_gen:
      TOK_P1    { $$ = "p1";}
    | TOK_ID    { $$ = "id";}
    | TOK_F     { $$ = "f";}
    | TOK_F1    { $$ = "f1";}
    | TOK_SEQ   { $$ = "seq";}
    | TOK_SEQ1  { $$ = "seq1";}
    | TOK_ISEQ  { $$ = "iseq";}
    | TOK_BAG   { $$ = "bag";}
;
 /* var_name: ident { $$ = $1; };
 leaving out '(' op_name ')' for now */

 /* decl_name: ident { $$ = $1; };
 leaving out '(' op_name ')' for now */

branches:
   branch { $$ = list1($1); }
 | branches TOK_VBAR branch { $$ = append($1,list1($3)); }
;
branch:
   ident  { $$ = cons_terminal("constant",list1($1)); }
 | ident TOK_LANGLE expr TOK_RANGLE
  { $$ = cons_terminal("constructor",cons($1,list1($3))); }
 | predicate  { $$ = cons_terminal("predicate",list1($1)); }
;
schema_exp:
   TOK_FORALL schema_text '@' schema_exp 
  { $$ = cons_terminal("forall",cons($2,list1($4))); } 
 | TOK_EXISTS schema_text '@' schema_exp 
  { $$ = cons_terminal("exists",cons($2,list1($4))); } 
 | TOK_EXISTS1 schema_text '@' schema_exp 
  { $$ = cons_terminal("exists1",cons($2,list1($4))); } 
 | schema_exp1 { $$ = $1; }
;
schema_exp1:
   '[' schema_text ']' { $$ = $2; } 
 | schema_or_var_ref { $$ = $1; } 
 | TOK_NOT schema_or_var_ref { $$ = cons_terminal("not",$2); } 
 | TOK_PRE schema_or_var_ref { $$ = cons_terminal("pre",$2); } 
 | schema_exp1 TOK_AND schema_exp1 { $$ = cons_terminal("and",cons($1,list1($3))); } 
 | schema_exp1 TOK_OR schema_exp1 { $$ = cons_terminal("or",cons($1,list1($3))); } 
 | schema_exp1 TOK_IMPLIES schema_exp1 { $$ = cons_terminal("implies",cons($1,list1($3))); } 
 | schema_exp1 TOK_IFF schema_exp1 { $$ = cons_terminal("iff",cons($1,list1($3))); } 
 | schema_exp1 TOK_PROJECT schema_exp1 { $$ = cons_terminal("project",cons($1,list1($3))); } 
 | schema_exp1 TOK_HIDE '(' decl_names ')' { $$ = cons_terminal("hide",cons($1,list1($4))); } 
 | schema_exp1 TOK_SEMI schema_exp1 { $$ = cons_terminal("semi",cons($1,list1($3))); } 
 | schema_exp1 TOK_RANGLE schema_exp1 { $$ = cons_terminal("rangle",cons($1,list1($3))); } 
 | '(' schema_exp ')' { $$ = $2; } 
;
schema_text:
   decl_part TOK_VBAR predicate { $$ = cons_terminal("schematext",cons($1,list1($3))); } 
 | decl_part { $$ = cons_terminal("schematext",cons($1,NULL)); } 
;
 /* schema_name: ident { $$ = $1; }; */

schema_or_var_ref:
   decorated_ident gen_actuals '[' renamings ']'
  { $$ = cons_terminal("ref",cons($1,cons($2,list1($4))));} 
 | decorated_ident gen_actuals
  { $$ = cons_terminal("ref",cons($1,cons($2,cons_terminal("nil",NULL)))); } 
 | decorated_ident '[' renamings ']'
  { $$ = cons_terminal("ref",cons($1,cons_terminal("nil",list1($3)))); } 
 | decorated_ident  { $$ = $1; } 
 | TOK_DELTA ident  { $$ = cons_terminal("delta",list1($2)); } 
 | TOK_XI ident     { $$ = cons_terminal("xi",list1($2)); } 
;
renamings:
   renaming { $$ = list1($1); }
 | renamings renaming { $$ = append($1,list1($2)); }
;
renaming:
 ident '/' ident { $$ = cons($3,list1($1));  }
;
basic_decl:
   decl_names ':' expr { $$ = cons_terminal("basic-decl",cons($3,$1)); } 
 | decl_names { $$ = cons_terminal("basic-decl",cons_terminal("nil",$1)); } 
 | schema_or_var_ref { $$ = $1; }
;
decl_names:
   ident { $$ = list1($1); } 
 | decl_names ',' ident { $$ = append($1,list1($3)); }
;
predicate:
   TOK_FORALL schema_text '@' predicate { $$ = cons_terminal("forall",cons($2,list1($4))); }  
 | TOK_EXISTS schema_text '@' predicate { $$ = cons_terminal("exists",cons($2,list1($4))); }  
 | TOK_EXISTS1 schema_text '@' predicate { $$ = cons_terminal("exists1",cons($2,list1($4))); }
 | TOK_LET let_defs '@' predicate { $$ = cons_terminal("let",cons($2,list1($4))); }
 | predicate1 { $$ = $1; }  
;
predicate1:
   expr { $$ = $1; }
 | predicate1 TOK_EQUAL expr { $$ = cons_terminal("equal",cons($1,list1($3))); }
 | predicate1 TOK_IN expr { $$ = cons_terminal("in",cons($1,list1($3))); }
 | predicate1 TOK_NEQ expr { $$ = cons_terminal("neq",cons($1,list1($3))); }
 | predicate1 TOK_NOTIN expr { $$ = cons_terminal("notin",cons($1,list1($3))); }
 | predicate1 TOK_SUB expr { $$ = cons_terminal("sub",cons($1,list1($3))); }
 | predicate1 TOK_SUBEQ expr { $$ = cons_terminal("subeq",cons($1,list1($3))); }
 | predicate1 TOK_LTE expr { $$ = cons_terminal("<=",cons($1,list1($3))); }
 | predicate1 '<' expr { $$ = cons_terminal("<",cons($1,list1($3))); }
 | predicate1 TOK_GTE expr { $$ = cons_terminal(">=",cons($1,list1($3))); }
 | predicate1 '>' expr { $$ = cons_terminal(">",cons($1,list1($3))); }
 | predicate1 TOK_PARTITION expr { $$ = cons_terminal("partition",cons($1,list1($3))); }
 | predicate1 TOK_INBAG expr { $$ = cons_terminal("inbag",cons($1,list1($3))); }
 | predicate1 TOK_INREL '{' ident '}' expr
  { $$ = cons(cons_terminal("underline",$4),cons($1,list1($6))); }
 | TOK_DISJOINT expr { $$ = cons_terminal("disjoint",list1($2)); }
 /* handled by expr | schema_or_var_ref { $$ = $1; } */
 | TOK_PRE schema_or_var_ref { $$ = cons_terminal("pre",list1($2)); }
 | TOK_TRUE  { $$ = make_name($1); }
 | TOK_FALSE { $$ = make_name($1); }
 | TOK_NOT predicate1 { $$ = cons_terminal("not",list1($2)); }
 | predicate1 TOK_AND predicate1 { $$ = cons_terminal("and",cons($1,list1($3))); }
 | predicate1 TOK_OR predicate1 { $$ = cons_terminal("or",cons($1,list1($3))); }
 | predicate1 TOK_IMPLIES predicate1 { $$ = cons_terminal("implies",cons($1,list1($3))); }
 | predicate1 TOK_IFF predicate1 { $$ = cons_terminal("iff",cons($1,list1($3))); }
 | '(' predicate ')' { $$ = $2; }
;
decorated_ident:
   ident deco_list { $$ = cons($1,list1($2)); }
 | ident { $$ = $1; }
;
ident: TOK_IDENTIFIER { $$ = cons_terminal(remove_tag, cons_symbol($1, NULL));}
;
idents:   ident         { $$ = list1($1);}
     | idents ',' ident { $$ = append($1, list1($3));}
;
let_defs:
     let_def  { $$ = list1($1);}
   | let_defs ';' let_def { $$ = append($1,list1($3));}
;
let_def:
     ident TOK_EQUIV expr { $$ = cons($1,list1($3));}
;
expr0:
      TOK_LAMBDA schema_text '@' expr  { $$ = cons_terminal("lambda", cons($2, list1($4)));}
    | TOK_UNIQUE schema_text '@' expr  { $$ = cons_terminal("unique", cons($2, list1($4)));}
    | TOK_UNIQUE schema_text           { $$ = cons_terminal("unique", cons($2, NULL));}
    | TOK_LET    let_defs    '@' expr  { $$ = cons_terminal("let", cons($2, list1($4)));}
    | expr                             { $$ = $1;}
;
expr:
      TOK_IF predicate TOK_THEN expr TOK_ELSE expr
     { $$ = cons_terminal("if", cons($2, cons($4,list1($6))));}
    | expr TOK_REL expr   { $$ = cons_terminal("rel", cons($1, list1($3)));}
    | expr TOK_PFUN expr   { $$ = cons_terminal("pfun", cons($1, list1($3)));}
    | expr TOK_TFUN expr   { $$ = cons_terminal("tfun", cons($1, list1($3)));}
    | expr TOK_PINJ expr   { $$ = cons_terminal("pinj", cons($1, list1($3)));}
    | expr TOK_INJ expr   { $$ = cons_terminal("inj", cons($1, list1($3)));}
    | expr TOK_PSURJ expr   { $$ = cons_terminal("psurj", cons($1, list1($3)));}
    | expr TOK_SURJ expr   { $$ = cons_terminal("surj", cons($1, list1($3)));}
    | expr TOK_BIJ expr   { $$ = cons_terminal("bij", cons($1, list1($3)));}
    | expr TOK_FFUN expr   { $$ = cons_terminal("ffun", cons($1, list1($3)));}
    | expr TOK_FINJ expr   { $$ = cons_terminal("finj", cons($1, list1($3)));}
    | expr TOK_HIDE expr   { $$ = cons_terminal("hide", cons($1, list1($3)));}
    | expr1 { $$ = $1; }
;
expr1:
    expr1 TOK_CROSS expr1 {
      if (stringp_car($3) && stringequal(strcar($3),"CROSS")) {
	$$ = cons_terminal("cross", cons($1, cdr($3))); }
      else $$ = cons_terminal("cross", cons($1, list1($3))); };

    | expr1 TOK_MAPSTO expr1   { $$ = cons_terminal("mapsto", cons($1, list1($3)));}
    | expr1 TOK_UPTO expr1   { $$ = cons_terminal("upto", cons($1, list1($3)));}
    | expr1 '+' expr1  { $$ = cons_terminal("+", cons ($1, list1($3)));} 
    | expr1 '-' expr1 { $$ = cons_terminal("-", cons ($1, list1($3)));} 
    | expr1 '*' expr1 { $$ = cons_terminal("*", cons ($1, list1($3)));}
    | expr1 '\\' expr1   { $$ = cons_terminal("/", cons ($1, list1($3)));}
    | expr1 TOK_CAT expr1   { $$ = cons_terminal("cat", cons ($1, list1($3)));}
    | expr1 TOK_UPLUS expr1   { $$ = cons_terminal("uplus", cons ($1, list1($3)));}
    | expr1 TOK_DIV expr1   { $$ = cons_terminal("div", cons ($1, list1($3)));}
    | expr1 TOK_MOD expr1   { $$ = cons_terminal("mod", cons ($1, list1($3)));}
    | expr1 TOK_CAP expr1   { $$ = cons_terminal("intersection", cons ($1, list1($3)));}
    | expr1 TOK_CUP expr1   { $$ = cons_terminal("union", cons ($1, list1($3)));}
    | expr1 TOK_COMP expr1   { $$ = cons_terminal("comp", cons ($1, list1($3)));}
    | expr1 TOK_CIRC expr1   { $$ = cons_terminal("circ", cons ($1, list1($3)));}
    | expr1 TOK_FILTER expr1   { $$ = cons_terminal("filter", cons ($1, list1($3)));}
    | expr1 TOK_OPLUS expr1   { $$ = cons_terminal("oplus", cons ($1, list1($3)));}
    | expr1 TOK_DRES expr1   { $$ = cons_terminal("dres", cons ($1, list1($3)));}
    | expr1 TOK_RRES expr1   { $$ = cons_terminal("rres", cons ($1, list1($3)));}
    | expr1 TOK_NRRES expr1   { $$ = cons_terminal("nrres", cons ($1, list1($3)));}
    | expr1 TOK_NDRES expr1   { $$ = cons_terminal("ndres", cons ($1, list1($3)));}
    | TOK_SIZE expr1         { $$ = cons_terminal("size",list1($2)); }
    | TOK_P expr3            { $$ = cons_terminal("powerset", list1($2));} 
    | TOK_DOM expr3            { $$ = cons_terminal("domain", list1($2));} 
    | TOK_RAN expr3            { $$ = cons_terminal("range", list1($2));} 
    | pre_gen expr3          { $$ = cons_terminal($1, list1($2));} 
    | '-' expr1              { $$ = cons_terminal("negate", list1($2));} 
    | expr2                  { $$ = $1;} 
;
expr2:
      expr3 expr2  { $$ = cons($1, list1($2)); }
    | expr3        { $$ = $1;}
;

/* expr2:
      expr2 expr3
     { if (!cdrp($1) ||
	   ($1->caflg == tstring && string_equal($1->pcar.atom, remove_tag))) 
          $$ = cons($1, list1($2));
       else $$ = append($1, list1($2)); }
    | expr3          { $$ = $1;}
; */

expr3:
    /* handled below - ident gen_actuals     { $$ = cons($1, list1($2));} */
      schema_or_var_ref        { $$ = $1;}
    | TOK_N                    { $$ = make_name("n");}
    | TOK_N1                   { $$ = make_name("n1");}
    | TOK_Z                    { $$ = make_name("z");}
    | integer_literal          { $$ = $1;}
    | TOK_EMPTY                { $$ = make_name("empty");}
    | set_expr                 { $$ = $1;}
    | TOK_LANGLE expr_list TOK_RANGLE { $$ = cons_terminal("list",$2);}
    | TOK_LANGLE expr0 '.' expr0 TOK_RANGLE
     { $$ = cons_terminal("list", cons_terminal("mapsto",cons($2,list1($4))));}
    | TOK_LBAG expr_list TOK_RBAG     { $$ = cons_terminal("bag",$2);}
    | '(' expr_list ')' { $$ = cons_terminal("tuple",$2);}
    | TOK_THETA ident    { $$ = cons_terminal("theta",list1($2));}
    | expr3 TOK_DOT ident   { $$ = cons_terminal("select",cons($1,list1($3)));}
    | expr3 TOK_INV          { $$ = cons_terminal("inv", list1($1));} 
    | expr3 TOK_PLUS         { $$ = cons_terminal("plus", list1($1));} 
    | expr3 TOK_STAR         { $$ = cons_terminal("star", list1($1));} 
    | expr3 TOK_CAT expr3    { $$ = cons_terminal("cat", cons($1,list1($3)));} 
    | expr3 TOK_LIMG expr0 TOK_RIMG   { $$ = cons_terminal("image", cons($1,list1($3)));} 
    | '(' expr0 ')'            { $$ = cons_terminal("parens", list1($2));}
;
expr_list: expr_list ',' expr { $$ = append($1, list1($3));}
     | expr          { $$ = list1($1);}
;
gen_actuals:
    '[' idents ']' { $$ = $2; }
;
gen_formals:
    '[' expr_list ']' { $$ = $2; }
;

set_expr:
      '{' '}' { $$ = cons_terminal("set*",NULL);}
    | '{' schema_text '@' expr '}' { $$ = cons_terminal("set-comp",cons($2,list1($4)));}
    | '{' schema_text '}'          { $$ = cons_terminal("set-comp",cons($2,NULL));}
    | '{' expr ',' expr_list '}'   { $$ = cons_terminal("set*",cons($2,$4));}
    | '{' expr '}'                 { $$ = cons_terminal("set*",cons($2,NULL));}
;

integer_literal: TOK_INTEGER_LITERAL
  { $$ = cons_terminal(remove_tag, cons_terminal($1,NULL)); }
;
string_literal: TOK_STRING_LITERAL
  { $$ = cons_terminal(remove_tag, cons_terminal($1,NULL)); }
;

%%

extern FILE  *yyin;

main() { 
  /* Remove the comments around yydebug = 1 to incorporate massive
     debugging information in your parse. */
  /*  yydebug = 1; */
 yyin = stdin ; /* fopen ("avaex1", "r"); */

 while (!feof(yyin)) { yyparse (); }
 }

/* %union{
 char *charptr;
 CELL  *treeptr;
 OBJCT *symptr;
 }
*/

/* token <symptr> IDENTIFIER */
