%{

#include "scansubs.h"
#include "semsubs.h"
#include "ztc-gram.tab.h"
#include <math.h>

%}

%p 2000
%a 1000
%e 1000
%o 1000

%%

 /*  Comments are handled by preprocessor (cpp). */

\n       { yylineno++; } /* yylineno++;*/

 /* In order to allow mixed case, use flex with the -i option. */

"and"    {return TOK_AND;}
"axiom"  {return TOK_AXIOM;}
"bag"    {return TOK_BAG;}

"begin"  { return TOK_BEGIN;}
"disjoint"  { return TOK_DISJOINT;}
"div"    { return TOK_DIV;}
"dom"    { return TOK_DOM;}
"if"     { return TOK_IF;}
"end"    { return TOK_END;}
"false"  { return TOK_FALSE;}

"forall" {return TOK_FORALL;}
"else"   { return TOK_ELSE;}
"exists" { return TOK_EXISTS;}
"exists1" { return TOK_EXISTS1;}

"generic" {return TOK_GENERIC;}     
"global"  {return TOK_GLOBAL;}     
"in"      {return TOK_IN;}
"inbag"   {return TOK_INBAG;}
"id"      {return TOK_ID;}

"is"      {return TOK_IS;}

"iseq"    {return TOK_ISEQ;}
"lambda"  {return TOK_LAMBDA;}
"let"     {return TOK_LET;}
"inrel"   {return TOK_INREL;}
"mod"     {return TOK_MOD;}
"not"     {return TOK_NOT;}
"notin"   {return TOK_NOTIN;}
"partition"  {return TOK_PARTITION;}
"pre"     {return TOK_PRE;}
"ran"     {return TOK_RAN;}
"or"      {return TOK_OR;}
"schema"  {return TOK_SCHEMA ; }
"seq"     {return TOK_SEQ ; }

"seq1"    {return TOK_SEQ1 ;}
"specification" {return TOK_SPEC; }
"initialize"    {return TOK_INIT; }
"spec"    {return TOK_SPEC; }
"data"    {return TOK_DATA; }
"sub"     {return TOK_SUB; }
"subeq"   {return TOK_SUBEQ; }
"then"    {return TOK_THEN; }
"theta"   {return TOK_THETA; }
"true"    {return TOK_TRUE; }
"unique"  {return TOK_UNIQUE; }
"where"   {return TOK_WHERE; }
"Delta"   {return TOK_DELTA; }
"F"       {return TOK_F; }
"FALSE"   {return TOK_FALSE; }
"F1"      {return TOK_F1; }
"Intersection" {return TOK_INTERSECTION; }
"N"       {return TOK_N; }
"N1"      {return TOK_N1; }
"P"       {return TOK_P; }
"P1"      {return TOK_P1; }
"TRUE"    {return TOK_TRUE; }
"Union"   {return TOK_UNION; }
"Xi"      {return TOK_XI; }
"Z"       {return TOK_Z; }                          

 /* Underscore not allowed in identifiers? ZTC */

[a-zA-Z][a-zA-Z0-9]*['?!]* { yylval.symptr = symlook (strdup(yytext), symtab, NSYMS);
                        return TOK_IDENTIFIER;}

 /* The following will fail to read "\"" as a string of length one */
 /* containing a  single dbl-quote. Z has no strings.              */

\"\"       { yylval.charptr = (char *)strdup(yytext);
             return TOK_STRING_LITERAL; }

\"[^"]+\"  { if (yytext[yyleng-2] == '\\')
                { yyless(yyleng-1);
                  yymore(); }    
              else { yylval.charptr = (char *)strdup(yytext);
                     return TOK_STRING_LITERAL; }; }

 /* No rational literals

 -?[1-9][0-9]*"/"-?[1-9][0-9]* { yylval.charptr = (char *)strdup(yytext);
                              return TOK_RATIONAL_LITERAL; }

 */

0   { yylval.charptr = (char *)strdup(yytext);
            return TOK_INTEGER_LITERAL; }

[1-9][0-9]* { yylval.charptr = (char *)strdup(yytext);
                return TOK_INTEGER_LITERAL; }

0[0-7]+ { yylval.charptr = (char *)strcat(strdup("#o"),yytext);
            return TOK_INTEGER_LITERAL; }


">->>"  {return TOK_BIJ;}
">++>"  {return TOK_FINJ;}

"::="  {return TOK_FREE_TYPE;} 

"<=>"  {return TOK_IFF;}
"<->"  {return TOK_REL;}
"-+>"  {return TOK_PFUN;}  /* tvfuzz */
"+->"  {return TOK_PFUN;}  /* ztc */
"-->"  {return TOK_TFUN;}
">+>"  {return TOK_PINJ;}
">->"  {return TOK_INJ;}
"+>>"  {return TOK_PSURJ;}
"->>"  {return TOK_SURJ;}
"++>"  {return TOK_FFUN;}

"<="  {return TOK_LTE;}
">="  {return TOK_GTE;}
"^+"  {return TOK_PLUS;}
"^*"  {return TOK_STAR;}

"||"  {return TOK_CUP;}
"&&"  {return TOK_CAP;}
"++"  {return TOK_UPLUS;}
":>"  {return TOK_COMP;}
"<:"  {return TOK_CIRC;}
"|-"  {return TOK_FILTER;}
"+="  {return TOK_OPLUS;}
"<|"  {return TOK_DRES;}
"|>"  {return TOK_RRES;}
"<+"  {return TOK_NDRES;}
"+>"  {return TOK_NRRES;}
"/="  {return TOK_NEQ;}

"=="  {return TOK_EQUIV;}
"<<"  {return TOK_LANGLE;}
">>"  {return TOK_RANGLE;}
"=>"  {return TOK_IMPLIES;}
"|\\"  {return TOK_PROJECT;}
"//"  {return TOK_SEMI;}
"(|"  {return TOK_LIMG;}
"|)"  {return TOK_RIMG;}
"[["  {return TOK_LBAG;}
"]]"  {return TOK_RBAG;}
"->"  {return TOK_MAPSTO;} 
".."  {return TOK_UPTO;} 
"{}"  {return TOK_EMPTY;}
"^/"  {return TOK_DCAT;}

"="  {return TOK_EQUAL;}
"&"  {return TOK_CROSS;}
"\\"  { return TOK_HIDE; }  /* SETMINUS */
"^"  { return TOK_CAT; }

"#"  {return TOK_SIZE;}

"|"  {return TOK_VBAR;}

"~"  { return TOK_INV; }

";"  {return TOK_SEP;}

[][:@+*<>?!'-.{}] {return yytext[0];}

 /* "\," ;
    "\!" ;
    "\:" ;
    "\;" ; */

[ \t\r]+ ;

%%

void yyerror (char *s)
{
 fail_flag = 1;
 fprintf (stderr, "\n;; *** %s at %s on line %d",s,yytext,yylineno+1);
};


